//@title Chuck-o-der
//@Author - Orchisama Das, Music 220A final project, 2016

//@descripton - A humble vocoder, that is built upon the works of Nick Gang and Gio Jacuzzi who
//first tried this in ChucK. This version is standalone and also adds a midi synthesizer component.
//A midi keyboard can be used to play polyphonic FM sounds. The magnitude spectrum of the microphone input determines 
//the spectrum of each synthesizer voice. The inverse fft of each synthesizer voice after processing sonifies the vocal processor.
//Midi controllers also control the FM parameters and can be adjusted to give a wide range of sounds. 
//I didn't use any effects, because I liked the raw sound from the keyboard and vocoder. Adding polyphony
//introduced a lot more harmonics to the processed vocal sound, making it more interesting.

1 => dac.gain;

// create our OSC receiver
OscIn oscin;
// a thing to retrieve message contents
OscMsg msg;
// use port 12000 (default Wekinator output port)
12000 => oscin.port;

// listen for "/wek/output" message with 8 floats coming in
oscin.addAddress( "/wek/outputs, ffffffff" );
// print
<<< "listening for OSC message from Wekinator on port 12000...", "" >>>;
<<< " |- expecting \"/wek/outputs\" with 8 continuous parameters...", "" >>>;

/*
 Ab2 Bb2 C3  Db3 Eb3 F3  G3  Ab3 */
[44, 46, 48, 49, 51, 53, 55, 56] @=> int root_scale[];
/*
 Eb3 F3  G3  Ab3 Bb3 C4  Db3 Eb4 */
[51, 53, 55, 56, 58, 60, 61, 63] @=> int voice2_scale[];
/*
 C4  Db4 Eb4 F4  G4  Ab4 Bb4 C5 */
[60, 61, 63, 65, 67, 68, 70, 72] @=> int voice3_scale[];

// [56, 58, 60, 61, 63, 65, 67, 68] @=> int scale_oct;

//connect mic input to FFT through a DC blocker
adc => PoleZero dcblock_in => FFT mic_fft => blackhole;
//connect inverse FFT through DC blocker and gain UGen to dac
PoleZero dcblock_out; 
Gain gain_out;

//set polezero position of DC blocker
0.999 => dcblock_in.blockZero;
0.999 => dcblock_out.blockZero;
0.8 => gain_out.gain;

// expecting 8 output dimensions
8 => int NUM_PARAMS;
float myParams[NUM_PARAMS];

// envelopes for smoothing parameters
// (alternately, can use slewing interpolators; SEE:
// https://chuck.stanford.edu/doc/examples/vector/interpolate.ck)
Envelope envs[NUM_PARAMS];
for( 0 => int i; i < NUM_PARAMS; i++ )
{
    envs[i] => blackhole;
    .5 => envs[i].value;
    10::ms => envs[i].duration;
}

//Polyphonic FM synth
3 => int maxVoices;
//change this line to whatever FM instrument you like
BeeThree synth[maxVoices];
FFT synth_fft[maxVoices];
IFFT out_ifft[maxVoices];

TriOsc fmosc => SawOsc voice1 => synth_fft[0] => blackhole;
fmosc => SawOsc voice2 => synth_fft[1] => blackhole;
fmosc => SawOsc voice3 => synth_fft[2] => blackhole;
2 => voice1.sync;
2 => voice2.sync;
2 => voice3.sync;

LPF lpf => NRev r => dac;
10000 => lpf.freq;
0.01 => r.mix;

/*adc => Gain adc_gain => LPF voice_lpf => HPF voice_hpf => dcblock_out => gain_out => lpf;
60 => voice_lpf.freq;
3000 => voice_hpf.freq;
0.1 => adc_gain.gain;*/

for(0 => int i; i< maxVoices; i++){
    // synth[i] => synth_fft[i] => blackhole;
    out_ifft[i] => dcblock_out => gain_out => lpf;
    //to also hear synth, uncomment the following lines
    // synth[i] => dac;
    // 0.1 => synth[i].gain;
}

int id[maxVoices];//an array to hold the note numbers so that we can match them up with the off signal
int  counter; 

public void playSynth(int note, int vel, int pos){
    //<<<note, vel>>>;
    Std.mtof(note) => synth[pos].freq;
    Math.log(vel)/2 => synth[pos].noteOn;
}

public void stopSynth(int pos){
    0 => synth[pos].noteOn;
}

public void changeFMDepth(int depth){
    for(0 => int i; i< maxVoices; i++)
        depth/128.0 => synth[i].lfoDepth;
}

public void changeFMModulation(int speed){
   for(0 => int i; i< maxVoices; i++)
      (speed/128.0) * 5 => synth[i].lfoSpeed;
}

// set the latest parameters as targets
// NOTE: we rely on map2sound() to actually interpret these parameters musically
fun void setParams( float params[] )
{
    // make sure we have enough
    if( params.size() >= NUM_PARAMS )
    {		
        // adjust the synthesis accordingly
        0.0 => float x;
        for( 0 => int i; i < NUM_PARAMS; i++ )
        {
            // get value
            params[i] => x;
            // clamp it
            if( x < 0 ) 0 => x;
            if( x > 1 ) 1 => x;
            // set as target of envelope (for smoothing)
            x => envs[i].target;
            // remember
            x => myParams[i];
        }
    }
}

// function to map incoming parameters to musical parameters
fun void map2sound()
{
    // time loop
    while( true )
    {
        /*// FYI envs[i] are used for smoothing param values
        envs[0].value() * 40 + 40 => Std.mtof => s.freq;
        // envs[0].value() * 40 + 37 => Std.mtof => s2.freq;
        envs[1].value() * 12000 + 50 => lpf.freq;
        envs[2].value() * .4 => s.gain;
        // envs[2].value() * .3 => s2.gain;
        envs[3].value() * 120 + 20 => Std.mtof => fmosc.freq;
        envs[4].value() * 400 => fmosc.gain;
        // envs[4].value() * 0.5 + 0.5 => breakbeat.gain;
        // envs[5].value() * 0.2 + 0.9 => breakbeat.rate;
        envs[6].value() * 0.15 => r.mix;*/
        
        (envs[0].value() * 7.99) $ int => int scale_index;
        envs[1].value() => float gain_switch;
        envs[2].value() => float octave_switch;
        (envs[3].value() * 7.99) $ int => int fm_index;
        envs[4].value() => float fm_octave_switch;
        envs[5].value() * 400 => fmosc.gain;  // TODO: TWEAK.
        Math.log10(1 + envs[6].value() * 9) * 20000 + 1000 => lpf.freq;
        envs[7].value() * 0.1 => r.mix;

        0 => int octave_num;
        if (octave_switch < 0.33)
        {
            -1 => octave_num;
        }
        else if (octave_switch >= 0.67)
        {
            1 => octave_num;
        }
        /*if (octave_switch < 0.2)
        {
            -2 => octave_num;
        }
        else if (octave_switch < 0.4)
        {
            -1 => octave_num;
        }
        else if (octave_switch >= 0.6)
        {
            1 => octave_num;
        }
        else if (octave_switch >= 0.8)
        {
            2 => octave_num;
        }*/

        root_scale[scale_index] + 12 * (octave_num + 1) => Std.mtof => voice1.freq;
        voice2_scale[scale_index] + 12 * (octave_num + 1) => Std.mtof => voice2.freq;
        voice3_scale[scale_index] + 12 * (octave_num + 1) => Std.mtof => voice3.freq;

        if (gain_switch < 0.33)
        {
            gain_switch * 3 + 0.01 => float zeroToOneMap;
            Math.log10(1 + zeroToOneMap * 9) => voice1.gain;
            0 => voice2.gain;
            0 => voice3.gain;

            1 => id[0];
            0 => id[1];
            0 => id[2];
        }
        else if (gain_switch < 0.67)
        {
            (gain_switch - 0.33) * 3 + 0.01 => float zeroToOneMap;
            1 => voice1.gain;
            Math.log10(1 + zeroToOneMap * 9) => voice2.gain;
            0 => voice3.gain;

            1 => id[0];
            1 => id[1];
            0 => id[2];
        }
        else
        {
            (gain_switch - 0.67) * 3 + 0.01 => float zeroToOneMap;
            1 => voice1.gain;
            1 => voice2.gain;
            Math.log10(1 + zeroToOneMap * 9) => voice3.gain;

            1 => id[0];
            1 => id[1];
            1 => id[2];
        }

        0 => int fm_octave_num;
        if (fm_octave_switch < 0.2)
        {
            -2 => fm_octave_num;
        }
        else if (fm_octave_switch < 0.4)
        {
            -1 => fm_octave_num;
        }
        else if (fm_octave_switch >= 0.6)
        {
            1 => fm_octave_num;
        }
        else if (fm_octave_switch >= 0.8)
        {
            2 => fm_octave_num;
        }

        root_scale[fm_index] + 12 * (fm_octave_num + 1) => Std.mtof => fmosc.freq;
        
        /*if (envs[2].value() > 0.7 && envs[2].value() < 0.85)
        {
            soundOff();
        }
        else
        {
            soundOn();
        }*/
        // time
        10::ms => now;
    }
}

fun void waitForEvent()
{
    // array to hold params
    float p[NUM_PARAMS];

    // infinite event loop
    while( true )
    {
        // wait for OSC message to arrive
        oscin => now;

        // grab the next message from the queue. 
        while( oscin.recv(msg) )
        {
            // print stuff
            cherr <= msg.address <= " ";
            
            // unpack our 5 floats into our array p
            for( int i; i < NUM_PARAMS; i++ )
            {
                // put into array
                msg.getFloat(i) => p[i];
                // print
                cherr <= p[i] <= " ";
            }
            
            // print
            cherr <= IO.newline();
            
            // set the parameters
            setParams( p );
        }
    }
}

fun void synthLoop(){
    
while( true ){
    200::ms => now;

    /* Math.random2(30,60) => int note;
    playSynth(note, Math.random2(10, 50), 0);
    playSynth(note + 7, Math.random2(10, 50), 1);
    playSynth(note + 16, Math.random2(10, 50), 2);*/
    /*30 => id[0];
    30 => id[1];
    30 => id[2];*/
    /*note + 7 => id[1];
    note + 16 => id[2];

    200::ms => now;

    stopSynth(0);
    stopSynth(1);
    stopSynth(2);*/
   /*// wait on your midi event
   10::ms => now;
    
	//this processes the keypresses
  while( min.recv( msg ) ){
   
     if( msg.data1 == 144){ //note on?
	  	0=> counter;         //initialize the counter
   	
	  	while(id[counter]!=0) //this looks for an empty position in the array
	   	    counter++;
	
	    msg.data2=>id[counter]; //puts the midi key number into the id array
		
        //call play function here
	  	playSynth(id[counter],msg.data3,counter);  
      }
        
      //use CC#1 and CC#2 to control FM parameters
      //the midi numbers are specific to the particular keyboard I am using
      else if(msg.data1 > 144){
          <<<msg.data1, msg.data2, msg.data3>>>;
          if(msg.data1 == 176 && msg.data2 == 71)
              changeFMModulation(msg.data3);
          else if(msg.data1 == 176 && msg.data2 == 72)
              changeFMDepth(msg.data3);
      }
          
 	
	  if( msg.data1 == 128){//note off?
	  	0=>counter;
  	
 	  while(id[counter]!=msg.data2)//looks for the counter position of the note that turned off
 	  	counter++;
      
	 stopSynth(counter);        //OFF!!!!!
     0=>id[counter];            //changes the id of the note just turned off to zero for re-use
      }          
     }*/
 }
 }
 
 fun void vocoder(){
     //the plan here is to calculate take the magnitude spectrum of each synth voice individually and change it according to the mic input
     
     //initialise fft size, window size, hop size
     512 => int FFT_SIZE => mic_fft.size;
     FFT_SIZE/2 => int WIN_SIZE;
     WIN_SIZE/2 => int HOP_SIZE;
     for(0 => int i; i < maxVoices; i++){
         FFT_SIZE => synth_fft[i].size => out_ifft[i].size;
         Windowing.hann(WIN_SIZE) => synth_fft[i].window => out_ifft[i].window;
     }
     
     Windowing.hann(WIN_SIZE) => mic_fft.window;
     complex spectrum_synth[WIN_SIZE];
     complex spectrum_mic[WIN_SIZE];
     polar temp_polar_synth;
     polar temp_polar_mic;
     
     
     while(true){
        
         //modulate magnitude spectrum of FMSynth with mic input
          mic_fft.upchuck();
          mic_fft.spectrum(spectrum_mic);
         
         
          for(0 => int i; i < maxVoices; i++){
             synth_fft[i].upchuck();
             synth_fft[i].spectrum(spectrum_synth);
             for(0 => int j; j < WIN_SIZE;j++){
                 spectrum_synth[j]$polar => temp_polar_synth;
                 spectrum_mic[j]$polar => temp_polar_mic;
                 
                 temp_polar_mic.mag => temp_polar_synth.mag;
                 temp_polar_synth$complex => spectrum_synth[j];
             }
             out_ifft[i].transform(spectrum_synth);
             //the next if statement is VERY important
             //if there are no midi messages associated with a voice, chuck gives it a 
             //default frequency of 220Hz, which overpowers all other voices, and what 
             //we hear is a constant vocoder sound not changing pitch when you change   	     //notes on the MIDI keyboard.
             if(id[i] == 0)
                 0 => out_ifft[i].gain;
             else
                 0.3 => out_ifft[i].gain;
         }
             HOP_SIZE :: samp => now;  
         
     }
 }
 

// spork osc receiver loop
spork ~ waitForEvent();
// spork mapping function
spork ~ map2sound();
// spork vocoding function
spork ~ vocoder();

while (true)
{
    1::second => now;
}